/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2021 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
  GNU General Public License for more details.
*/

#ifndef UTIL_WILDCARDS_H
#define UTIL_WILDCARDS_H

#include <vector>
#include <string>

char *expand_filename(const char *args);

std::vector<std::string> expand_wild_cards(std::vector<std::string> argv);

int wildcardmatch(const char *w, const char *s);
int wildcardmatch(const std::string &w, const char *s);

#endif
