#line 1 "expr_lex.cc"

#line 3 "expr_lex.cc"

#define YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 6
#define YY_FLEX_SUBMINOR_VERSION 4
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

#ifdef yyget_lval
#define yyget_lval_ALREADY_DEFINED
#else
#define yyget_lval yyget_lval
#endif

#ifdef yyset_lval
#define yyset_lval_ALREADY_DEFINED
#else
#define yyset_lval yyset_lval
#endif

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

/* end standard C headers. */

/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types.
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <inttypes.h>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t;
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN (-32767 - 1)
#endif
#ifndef INT32_MIN
#define INT32_MIN (-2147483647 - 1)
#endif
#ifndef INT8_MAX
#define INT8_MAX (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX (4294967295U)
#endif

#ifndef SIZE_MAX
#define SIZE_MAX (~(size_t) 0)
#endif

#endif /* ! C99 */

#endif /* ! FLEXINT_H */

/* begin standard C++ headers. */

/* TODO: this is always defined, so inline it */
#define yyconst const

#if defined(__GNUC__) && __GNUC__ >= 3
#define yynoreturn __attribute__((__noreturn__))
#else
#define yynoreturn
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an
 *   integer in range [0..255] for use as an array index.
 */
#define YY_SC_TO_UI(c) ((YY_CHAR)(c))

/* An opaque pointer. */
#ifndef YY_TYPEDEF_YY_SCANNER_T
#define YY_TYPEDEF_YY_SCANNER_T
typedef void *yyscan_t;
#endif

/* For convenience, these vars (plus the bison vars far below)
   are macros in the reentrant scanner. */
#define yyin yyg->yyin_r
#define yyout yyg->yyout_r
#define yyextra yyg->yyextra_r
#define yyleng yyg->yyleng_r
#define yytext yyg->yytext_r
#define yylineno (YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno)
#define yycolumn (YY_CURRENT_BUFFER_LVALUE->yy_bs_column)
#define yy_flex_debug yyg->yy_flex_debug_r

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN yyg->yy_start = 1 + 2 *
/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START ((yyg->yy_start - 1) / 2)
#define YYSTATE YY_START
/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)
/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yyrestart(yyin, yyscanner)
#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k.
 * Moreover, YY_BUF_SIZE is 2*YY_READ_BUF_SIZE in the general case.
 * Ditto for the __ia64__ case accordingly.
 */
#define YY_BUF_SIZE 32768
#else
#define YY_BUF_SIZE 16384
#endif /* __ia64__ */
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

#define YY_LESS_LINENO(n)
#define YY_LINENO_REWIND_TO(ptr)

/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n)                                                       \
  do                                                                    \
    {                                                                   \
      /* Undo effects of setting up yytext. */                          \
      int yyless_macro_arg = (n);                                       \
      YY_LESS_LINENO(yyless_macro_arg);                                 \
      *yy_cp = yyg->yy_hold_char;                                       \
      YY_RESTORE_YY_MORE_OFFSET                                         \
      yyg->yy_c_buf_p = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
      YY_DO_BEFORE_ACTION; /* set up yytext again */                    \
    }                                                                   \
  while (0)
#define unput(c) yyunput(c, yyg->yytext_ptr, yyscanner)

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
{
  FILE *yy_input_file;

  char *yy_ch_buf;  /* input buffer */
  char *yy_buf_pos; /* current position in input buffer */

  /* Size of input buffer in bytes, not including room for EOB
   * characters.
   */
  int yy_buf_size;

  /* Number of characters read into yy_ch_buf, not including EOB
   * characters.
   */
  int yy_n_chars;

  /* Whether we "own" the buffer - i.e., we know we created it,
   * and can realloc() it to grow it, and should free() it to
   * delete it.
   */
  int yy_is_our_buffer;

  /* Whether this is an "interactive" input source; if so, and
   * if we're using stdio for input, then we want to use getc()
   * instead of fread(), to make sure we stop fetching input after
   * each newline.
   */
  int yy_is_interactive;

  /* Whether we're considered to be at the beginning of a line.
   * If so, '^' rules will be active on the next match, otherwise
   * not.
   */
  int yy_at_bol;

  int yy_bs_lineno; /**< The line count. */
  int yy_bs_column; /**< The column count. */

  /* Whether to try to fill the input buffer when we reach the
   * end of it.
   */
  int yy_fill_buffer;

  int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
  /* When an EOF's been seen but there's still some text to process
   * then we mark the buffer as YY_EOF_PENDING, to indicate that we
   * shouldn't try reading from the input source any more.  We might
   * still have a bunch of tokens to match, though, because of
   * possible backing-up.
   *
   * When we actually see the EOF, we change the status to "new"
   * (via yyrestart()), so that the user can continue scanning by
   * just pointing yyin at a new input file.
   */
#define YY_BUFFER_EOF_PENDING 2
};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER (yyg->yy_buffer_stack ? yyg->yy_buffer_stack[yyg->yy_buffer_stack_top] : NULL)
/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE yyg->yy_buffer_stack[yyg->yy_buffer_stack_top]

void yyrestart(FILE *input_file, yyscan_t yyscanner);
void yy_switch_to_buffer(YY_BUFFER_STATE new_buffer, yyscan_t yyscanner);
YY_BUFFER_STATE yy_create_buffer(FILE *file, int size, yyscan_t yyscanner);
void yy_delete_buffer(YY_BUFFER_STATE b, yyscan_t yyscanner);
void yy_flush_buffer(YY_BUFFER_STATE b, yyscan_t yyscanner);
void yypush_buffer_state(YY_BUFFER_STATE new_buffer, yyscan_t yyscanner);
void yypop_buffer_state(yyscan_t yyscanner);

static void yyensure_buffer_stack(yyscan_t yyscanner);
static void yy_load_buffer_state(yyscan_t yyscanner);
static void yy_init_buffer(YY_BUFFER_STATE b, FILE *file, yyscan_t yyscanner);
#define YY_FLUSH_BUFFER yy_flush_buffer(YY_CURRENT_BUFFER, yyscanner)

YY_BUFFER_STATE yy_scan_buffer(char *base, yy_size_t size, yyscan_t yyscanner);
YY_BUFFER_STATE yy_scan_string(const char *yy_str, yyscan_t yyscanner);
YY_BUFFER_STATE yy_scan_bytes(const char *bytes, int len, yyscan_t yyscanner);

void *yyalloc(yy_size_t, yyscan_t yyscanner);
void *yyrealloc(void *, yy_size_t, yyscan_t yyscanner);
void yyfree(void *, yyscan_t yyscanner);

#define yy_new_buffer yy_create_buffer
#define yy_set_interactive(is_interactive)                                         \
  {                                                                                \
    if (!YY_CURRENT_BUFFER)                                                        \
      {                                                                            \
        yyensure_buffer_stack(yyscanner);                                          \
        YY_CURRENT_BUFFER_LVALUE = yy_create_buffer(yyin, YY_BUF_SIZE, yyscanner); \
      }                                                                            \
    YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive;                  \
  }
#define yy_set_bol(at_bol)                                                         \
  {                                                                                \
    if (!YY_CURRENT_BUFFER)                                                        \
      {                                                                            \
        yyensure_buffer_stack(yyscanner);                                          \
        YY_CURRENT_BUFFER_LVALUE = yy_create_buffer(yyin, YY_BUF_SIZE, yyscanner); \
      }                                                                            \
    YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol;                                  \
  }
#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* Begin user sect3 */

#define yywrap(yyscanner) (/*CONSTCOND*/ 1)
#define YY_SKIP_YYWRAP
typedef flex_uint8_t YY_CHAR;

typedef int yy_state_type;

#define yytext_ptr yytext_r

static yy_state_type yy_get_previous_state(yyscan_t yyscanner);
static yy_state_type yy_try_NUL_trans(yy_state_type current_state, yyscan_t yyscanner);
static int yy_get_next_buffer(yyscan_t yyscanner);
static void yynoreturn yy_fatal_error(const char *msg, yyscan_t yyscanner);

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION       \
  yyg->yytext_ptr = yy_bp;        \
  yyleng = (int) (yy_cp - yy_bp); \
  yyg->yy_hold_char = *yy_cp;     \
  *yy_cp = '\0';                  \
  yyg->yy_c_buf_p = yy_cp;
#define YY_NUM_RULES 26
#define YY_END_OF_BUFFER 27
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
{
  flex_int32_t yy_verify;
  flex_int32_t yy_nxt;
};
static const flex_int16_t yy_acclist[149]
    = { 0,     4,  4,     27,    25, 26,    24,    25,    26, 24, 26,    21,    25, 26,    1,  25,    26,    25, 26,    11,    25,
        26,    4,  11,    25,    26, 4,     25,    26,    23, 25, 26,    16,    25, 26,    11, 25,    26,    15, 25,    26,    22,
        25,    26, 9,     10,    25, 26,    16391, 9,     10, 25, 26,    16391, 9,  10,    25, 26,    16391, 9,  10,    25,    26,
        16391, 9,  10,    25,    26, 16391, 25,    26,    24, 18, 1,     19,    4,  4,     4,  10,    10,    14, 17,    13,    8199,
        8,     9,  10,    16391, 9,  10,    16391, 4,     9,  10, 16391, 9,     10, 16391, 9,  10,    16391,

        9,     10, 16391, 20,    4,  10,    10,    4,     10, 12, 8,     4,     3,  9,     10, 16391, 9,     10, 16391, 9,     10,
        16391, 9,  10,    16391, 2,  9,     10,    16391, 9,  10, 16391, 9,     10, 16391, 9,  10,    16391, 9,  10,    16391, 6,
        8199,  9,  10,    16391, 5,  8199 };

static const flex_int16_t yy_accept[70]
    = { 0,   1,   2,   3,   4,   6,   9,   11,  14,  17,  19,  22,  26,  29,  32,  35,  38,  41,  44,  49,  54,  59, 64,  69,
        71,  72,  73,  74,  75,  76,  76,  77,  78,  79,  80,  81,  82,  83,  83,  84,  85,  88,  91,  91,  95,  98, 101, 104,
        105, 105, 106, 107, 108, 110, 111, 112, 113, 117, 120, 123, 126, 130, 133, 136, 139, 142, 144, 147, 149, 149 };

static const YY_CHAR yy_ec[256]
    = { 0,  1,  1,  1,  1,  1,  1,  1,  1,  2,  3,  1,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        1,  1,  1,  1,  4,  5,  1,  6,  1,  1,  7,  1,  8,  9,  9,  10, 9,  10, 11, 9,  12, 12, 12, 12, 12, 12, 12, 12,
        12, 12, 13, 9,  14, 15, 16, 17, 1,  18, 18, 18, 18, 19, 18, 18, 18, 20, 18, 18, 18, 21, 18, 18, 22, 18, 18, 18,
        18, 18, 18, 18, 18, 18, 18, 1,  1,  1,  9,  23, 1,  18, 18, 18, 18,

        24, 18, 18, 18, 25, 18, 18, 18, 26, 27, 28, 29, 18, 30, 18, 31, 18, 32, 18, 18, 18, 18, 9,  33, 9,  1,  1,  1,
        1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,

        1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 };

static const YY_CHAR yy_meta[34]
    = { 0, 1, 1, 2, 3, 1, 1, 1, 3, 1, 1, 4, 5, 1, 1, 1, 1, 1, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 1 };

static const flex_int16_t yy_base[74]
    = { 0,   0,  0,   157, 158, 32, 35, 141, 0,  148, 158, 28,  30,  158, 139, 138, 124, 158, 47,  56,  33,  51,  38,  103, 67,
        158, 0,  158, 53,  34,  61, 63, 123, 66, 118, 158, 158, 49,  158, 0,   74,  78,  120, 121, 79,  82,  83,  158, 117, 115,
        113, 96, 84,  158, 0,   71, 87, 90,  93, 94,  101, 102, 105, 111, 106, 158, 116, 158, 158, 137, 140, 144, 146, 73 };

static const flex_int16_t yy_def[74]
    = { 0,  68, 1,  68, 68, 68, 68, 68, 69, 68, 68, 68, 70, 68, 68, 68, 68, 68, 71, 71, 19, 19, 19, 68, 68,
        68, 69, 68, 68, 68, 68, 70, 72, 72, 68, 68, 68, 68, 68, 73, 19, 19, 68, 19, 19, 19, 19, 68, 68, 68,
        72, 72, 72, 68, 73, 68, 19, 19, 19, 19, 19, 19, 19, 19, 19, 68, 19, 68, 0,  68, 68, 68, 68, 68 };

static const flex_int16_t yy_nxt[192]
    = { 0,  4,  5,  6,  5,  7,  8,  9,  10, 10, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 18, 20, 18, 18, 19, 18, 18, 18, 18, 21, 22,
        18, 18, 23, 24, 24, 24, 24, 24, 24, 28, 30, 31, 68, 48, 40, 49, 29, 68, 33, 40, 37, 29, 37, 33, 38, 44, 38, 39, 40, 37, 68,
        46, 40, 38, 28, 42, 39, 43, 24, 24, 24, 29, 28, 30, 31, 42, 29, 52, 54, 29, 45, 33, 55, 68, 29, 40, 33, 68, 68, 40, 40, 68,
        68, 40, 40, 52, 68, 56, 40, 68,

        57, 40, 68, 68, 40, 40, 58, 50, 59, 60, 68, 68, 40, 40, 68, 68, 40, 40, 65, 61, 68, 62, 40, 67, 50, 68, 49, 40, 49, 66, 68,
        55, 63, 53, 50, 47, 64, 26, 36, 26, 26, 26, 26, 32, 32, 32, 41, 41, 41, 41, 51, 51, 35, 34, 27, 25, 68, 3,  68, 68, 68, 68,
        68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68

      };

static const flex_int16_t yy_chk[192]
    = { 0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        1,  1,  1,  5,  5,  5,  6,  6,  6,  11, 12, 12, 20, 29, 20, 29, 11, 22, 12, 22, 18, 11, 37, 12, 18, 20, 37, 18, 18, 19, 21,
        22, 21, 19, 28, 19, 19, 19, 24, 24, 24, 28, 30, 31, 31, 33, 28, 33, 73, 30, 21, 31, 55, 40, 30, 40, 31, 41, 44, 41, 44, 45,
        46, 45, 46, 52, 56, 44, 56, 57,

        44, 57, 58, 59, 58, 59, 45, 51, 46, 57, 60, 61, 60, 61, 62, 64, 62, 64, 63, 58, 63, 59, 63, 66, 50, 66, 49, 66, 48, 64, 43,
        42, 61, 34, 32, 23, 62, 69, 16, 69, 69, 69, 69, 70, 70, 70, 71, 71, 71, 71, 72, 72, 15, 14, 9,  7,  3,  68, 68, 68, 68, 68,
        68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68, 68

      };

#define YY_TRAILING_MASK 0x2000
#define YY_TRAILING_HEAD_MASK 0x4000
#define REJECT                                                                      \
  {                                                                                 \
    *yy_cp = yyg->yy_hold_char;             /* undo effects of setting up yytext */ \
    yy_cp = yyg->yy_full_match;             /* restore poss. backed-over text */    \
    yyg->yy_lp = yyg->yy_full_lp;           /* restore orig. accepting pos. */      \
    yyg->yy_state_ptr = yyg->yy_full_state; /* restore orig. state */               \
    yy_current_state = *yyg->yy_state_ptr;  /* restore curr. state */               \
    ++yyg->yy_lp;                                                                   \
    goto find_rule;                                                                 \
  }

#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
#line 1 "expr_lex.l"
/* flex -oexpr_lex.c expr_lex.l */
#line 3 "expr_lex.l"
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifndef M_E
#define M_E 2.7182818284590452354 /* e */
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846 /* pi */
#endif

#include "expr.h"
#include "expr_yacc.h"

#line 544 "expr_lex.cc"
/* Definitions:  LMB92 p. 153
   Definitions are named regular expressions, e.g., DGT [0-9]
   Definitions enclosed in braces in rules section, e.g. {DGT}, are interpreted literally
   DGT [0-9] Digit
   LPH [A-Za-z_] Alphabetic character
   LPHDGT [A-Za-z0-9_] Alphanumeric character
   XPN [eE][+-]?[0-9]+ Real number Exponent */
#line 552 "expr_lex.cc"

#define INITIAL 0

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
#include <unistd.h>
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

/* Holds the entire state of the reentrant scanner. */
struct yyguts_t
{

  /* User-defined. Not touched by flex. */
  YY_EXTRA_TYPE yyextra_r;

  /* The rest are the same as the globals declared in the non-reentrant scanner. */
  FILE *yyin_r, *yyout_r;
  size_t yy_buffer_stack_top;       /**< index of top of stack. */
  size_t yy_buffer_stack_max;       /**< capacity of stack. */
  YY_BUFFER_STATE *yy_buffer_stack; /**< Stack as an array. */
  char yy_hold_char;
  int yy_n_chars;
  int yyleng_r;
  char *yy_c_buf_p;
  int yy_init;
  int yy_start;
  int yy_did_buffer_switch_on_eof;
  int yy_start_stack_ptr;
  int yy_start_stack_depth;
  int *yy_start_stack;
  yy_state_type yy_last_accepting_state;
  char *yy_last_accepting_cpos;

  int yylineno_r;
  int yy_flex_debug_r;

  yy_state_type *yy_state_buf;
  yy_state_type *yy_state_ptr;
  char *yy_full_match;
  int yy_lp;

  /* These are only needed for trailing context rules,
   * but there's no conditional variable for that yet. */
  int yy_looking_for_trail_begin;
  int yy_full_lp;
  int *yy_full_state;

  char *yytext_r;
  int yy_more_flag;
  int yy_more_len;

  YYSTYPE *yylval_r;

}; /* end struct yyguts_t */

static int yy_init_globals(yyscan_t yyscanner);

/* This must go here because YYSTYPE and YYLTYPE are included
 * from bison output in section 1.*/
#define yylval yyg->yylval_r

int yylex_init(yyscan_t *scanner);

int yylex_init_extra(YY_EXTRA_TYPE user_defined, yyscan_t *scanner);

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int yylex_destroy(yyscan_t yyscanner);

int yyget_debug(yyscan_t yyscanner);

void yyset_debug(int debug_flag, yyscan_t yyscanner);

YY_EXTRA_TYPE yyget_extra(yyscan_t yyscanner);

void yyset_extra(YY_EXTRA_TYPE user_defined, yyscan_t yyscanner);

FILE *yyget_in(yyscan_t yyscanner);

void yyset_in(FILE *_in_str, yyscan_t yyscanner);

FILE *yyget_out(yyscan_t yyscanner);

void yyset_out(FILE *_out_str, yyscan_t yyscanner);

int yyget_leng(yyscan_t yyscanner);

char *yyget_text(yyscan_t yyscanner);

int yyget_lineno(yyscan_t yyscanner);

void yyset_lineno(int _line_number, yyscan_t yyscanner);

int yyget_column(yyscan_t yyscanner);

void yyset_column(int _column_no, yyscan_t yyscanner);

YYSTYPE *yyget_lval(yyscan_t yyscanner);

void yyset_lval(YYSTYPE *yylval_param, yyscan_t yyscanner);

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yywrap(yyscan_t yyscanner);
#else
extern int yywrap(yyscan_t yyscanner);
#endif
#endif

#ifndef YY_NO_UNPUT

#endif

#ifndef yytext_ptr
static void yy_flex_strncpy(char *, const char *, int, yyscan_t yyscanner);
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen(const char *, yyscan_t yyscanner);
#endif

#ifndef YY_NO_INPUT
#ifdef __cplusplus
static int yyinput(yyscan_t yyscanner);
#else
static int input(yyscan_t yyscanner);
#endif

#endif

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k */
#define YY_READ_BUF_SIZE 16384
#else
#define YY_READ_BUF_SIZE 8192
#endif /* __ia64__ */
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO                                         \
  do                                                 \
    {                                                \
      if (fwrite(yytext, (size_t) yyleng, 1, yyout)) \
        {                                            \
        }                                            \
    }                                                \
  while (0)
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf, result, max_size)                                                         \
  if (YY_CURRENT_BUFFER_LVALUE->yy_is_interactive)                                              \
    {                                                                                           \
      int c = '*';                                                                              \
      int n;                                                                                    \
      for (n = 0; n < max_size && (c = getc(yyin)) != EOF && c != '\n'; ++n) buf[n] = (char) c; \
      if (c == '\n') buf[n++] = (char) c;                                                       \
      if (c == EOF && ferror(yyin)) YY_FATAL_ERROR("input in flex scanner failed");             \
      result = n;                                                                               \
    }                                                                                           \
  else                                                                                          \
    {                                                                                           \
      errno = 0;                                                                                \
      while ((result = (int) fread(buf, 1, (yy_size_t) max_size, yyin)) == 0 && ferror(yyin))   \
        {                                                                                       \
          if (errno != EINTR)                                                                   \
            {                                                                                   \
              YY_FATAL_ERROR("input in flex scanner failed");                                   \
              break;                                                                            \
            }                                                                                   \
          errno = 0;                                                                            \
          clearerr(yyin);                                                                       \
        }                                                                                       \
    }

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error(msg, yyscanner)
#endif

/* end tables serialization structures and prototypes */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1

extern int yylex(YYSTYPE *yylval_param, yyscan_t yyscanner);

#define YY_DECL int yylex(YYSTYPE *yylval_param, yyscan_t yyscanner)
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK /*LINTED*/ break;
#endif

#define YY_RULE_SETUP YY_USER_ACTION

/** The main scanner function which does all the work.
 */
YY_DECL
{
  yy_state_type yy_current_state;
  char *yy_cp, *yy_bp;
  int yy_act;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  yylval = yylval_param;

  if (!yyg->yy_init)
    {
      yyg->yy_init = 1;

#ifdef YY_USER_INIT
      YY_USER_INIT;
#endif

      /* Create the reject buffer large enough to save one state per allowed character. */
      if (!yyg->yy_state_buf) yyg->yy_state_buf = (yy_state_type *) yyalloc(YY_STATE_BUF_SIZE, yyscanner);
      if (!yyg->yy_state_buf) YY_FATAL_ERROR("out of dynamic memory in yylex()");

      if (!yyg->yy_start) yyg->yy_start = 1; /* first start state */

      if (!yyin) yyin = stdin;

      if (!yyout) yyout = stdout;

      if (!YY_CURRENT_BUFFER)
        {
          yyensure_buffer_stack(yyscanner);
          YY_CURRENT_BUFFER_LVALUE = yy_create_buffer(yyin, YY_BUF_SIZE, yyscanner);
        }

      yy_load_buffer_state(yyscanner);
    }

  {
#line 36 "expr_lex.l"

#line 844 "expr_lex.cc"

    while (/*CONSTCOND*/ 1) /* loops until end-of-file is reached */
      {
        yy_cp = yyg->yy_c_buf_p;

        /* Support of yytext. */
        *yy_cp = yyg->yy_hold_char;

        /* yy_bp points to the position in yy_ch_buf of the start of
         * the current run.
         */
        yy_bp = yy_cp;

        yy_current_state = yyg->yy_start;

        yyg->yy_state_ptr = yyg->yy_state_buf;
        *yyg->yy_state_ptr++ = yy_current_state;

      yy_match:
        do
          {
            YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)];
            while (yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state)
              {
                yy_current_state = (int) yy_def[yy_current_state];
                if (yy_current_state >= 69) yy_c = yy_meta[yy_c];
              }
            yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
            *yyg->yy_state_ptr++ = yy_current_state;
            ++yy_cp;
          }
        while (yy_base[yy_current_state] != 158);

      yy_find_action:
        yy_current_state = *--yyg->yy_state_ptr;
        yyg->yy_lp = yy_accept[yy_current_state];
      find_rule: /* we branch to this label when backing up */
        for (;;) /* until we find what rule we matched */
          {
            if (yyg->yy_lp && yyg->yy_lp < yy_accept[yy_current_state + 1])
              {
                yy_act = yy_acclist[yyg->yy_lp];
                if (yy_act & YY_TRAILING_HEAD_MASK || yyg->yy_looking_for_trail_begin)
                  {
                    if (yy_act == yyg->yy_looking_for_trail_begin)
                      {
                        yyg->yy_looking_for_trail_begin = 0;
                        yy_act &= ~YY_TRAILING_HEAD_MASK;
                        break;
                      }
                  }
                else if (yy_act & YY_TRAILING_MASK)
                  {
                    yyg->yy_looking_for_trail_begin = yy_act & ~YY_TRAILING_MASK;
                    yyg->yy_looking_for_trail_begin |= YY_TRAILING_HEAD_MASK;
                  }
                else
                  {
                    yyg->yy_full_match = yy_cp;
                    yyg->yy_full_state = yyg->yy_state_ptr;
                    yyg->yy_full_lp = yyg->yy_lp;
                    break;
                  }
                ++yyg->yy_lp;
                goto find_rule;
              }
            --yy_cp;
            yy_current_state = *--yyg->yy_state_ptr;
            yyg->yy_lp = yy_accept[yy_current_state];
          }

        YY_DO_BEFORE_ACTION;

      do_action: /* This label is used only to access EOF actions. */

        switch (yy_act)
          { /* beginning of action switch */
          case 1:
            YY_RULE_SETUP
#line 38 "expr_lex.l"
                ; /* ignore comments */
            YY_BREAK
          case 2: YY_RULE_SETUP
#line 40 "expr_lex.l"
            {
              yylval->cvalue = M_PI;
              return CONSTANT;
            }
            YY_BREAK
          case 3: YY_RULE_SETUP
#line 45 "expr_lex.l"
            {
              yylval->cvalue = M_E;
              return CONSTANT;
            }
            YY_BREAK
          case 4: YY_RULE_SETUP
#line 50 "expr_lex.l"
            {
              /* constant */
              /* NB: Tempted to prepend lexer expressions for floats and doubles with [+-]?
               so that unary plus/minus is handled in lexer rather than parser.
               However, this has unintended side effects so let parser handle it for now */
              yylval->cvalue = strtod(yytext, (char **) NULL);
              return CONSTANT;
            } /* end constant */
            YY_BREAK
          case 5: YY_RULE_SETUP
#line 59 "expr_lex.l"
            {
              return REMOVE;
            }
            YY_BREAK
          case 6: YY_RULE_SETUP
#line 61 "expr_lex.l"
            {
              return PRINT;
            }
            YY_BREAK
          case 7: YY_RULE_SETUP
#line 63 "expr_lex.l"
            {
              yylval->fname = strdup(yytext);
              return FUNCTION;
            } /* end functions */
            YY_BREAK
          case 8: YY_RULE_SETUP
#line 68 "expr_lex.l"
            {
              /* Recognize coordinates, e.g., var_nm.x (lon of var_nm) */
              yylval->varnm = strdup(yytext);
              return VARIABLE;
            }
            YY_BREAK
          case 9: YY_RULE_SETUP
#line 74 "expr_lex.l"
            {
              yylval->varnm = strdup(yytext);
              return VARIABLE;
            }
            YY_BREAK
          case 10: YY_RULE_SETUP
#line 79 "expr_lex.l"
            {
              yylval->varnm = strdup(yytext);
              return VARIABLE;
            }
            YY_BREAK
          case 11: YY_RULE_SETUP
#line 85 "expr_lex.l"
            {
              return *yytext;
            }
            YY_BREAK
          case 12: YY_RULE_SETUP
#line 89 "expr_lex.l"
            return LEG;
            YY_BREAK
          case 13: YY_RULE_SETUP
#line 90 "expr_lex.l"
            return GE;
            YY_BREAK
          case 14: YY_RULE_SETUP
#line 91 "expr_lex.l"
            return LE;
            YY_BREAK
          case 15: YY_RULE_SETUP
#line 92 "expr_lex.l"
            return GT;
            YY_BREAK
          case 16: YY_RULE_SETUP
#line 93 "expr_lex.l"
            return LT;
            YY_BREAK
          case 17: YY_RULE_SETUP
#line 94 "expr_lex.l"
            return EQ;
            YY_BREAK
          case 18: YY_RULE_SETUP
#line 95 "expr_lex.l"
            return NE;
            YY_BREAK
          case 19: YY_RULE_SETUP
#line 96 "expr_lex.l"
            return AND;
            YY_BREAK
          case 20: YY_RULE_SETUP
#line 97 "expr_lex.l"
            return OR;
            YY_BREAK
          case 21: YY_RULE_SETUP
#line 98 "expr_lex.l"
            return NOT;
            YY_BREAK
          case 22: YY_RULE_SETUP
#line 99 "expr_lex.l"
            return QUESTION;
            YY_BREAK
          case 23: YY_RULE_SETUP
#line 100 "expr_lex.l"
            return COLON;
            YY_BREAK
          case 24:
            /* rule 24 can match eol */
            YY_RULE_SETUP
#line 102 "expr_lex.l"
                ; /* ignore whitespace */
            YY_BREAK
          case 25: YY_RULE_SETUP
#line 104 "expr_lex.l"
            yyerror(NULL, NULL, "Unknown character");
            YY_BREAK
          case 26: YY_RULE_SETUP
#line 105 "expr_lex.l"
            ECHO;
            YY_BREAK
#line 1083 "expr_lex.cc"
          case YY_STATE_EOF(INITIAL): yyterminate();

          case YY_END_OF_BUFFER:
            {
              /* Amount of text matched not including the EOB char. */
              int yy_amount_of_matched_text = (int) (yy_cp - yyg->yytext_ptr) - 1;

              /* Undo the effects of YY_DO_BEFORE_ACTION. */
              *yy_cp = yyg->yy_hold_char;
              YY_RESTORE_YY_MORE_OFFSET

              if (YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW)
                {
                  /* We're scanning a new file or input source.  It's
                   * possible that this happened because the user
                   * just pointed yyin at a new source and called
                   * yylex().  If so, then we have to assure
                   * consistency between YY_CURRENT_BUFFER and our
                   * globals.  Here is the right place to do so, because
                   * this is the first action (other than possibly a
                   * back-up) that will match for the new input source.
                   */
                  yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
                  YY_CURRENT_BUFFER_LVALUE->yy_input_file = yyin;
                  YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
                }

              /* Note that here we test for yy_c_buf_p "<=" to the position
               * of the first EOB in the buffer, since yy_c_buf_p will
               * already have been incremented past the NUL character
               * (since all states make transitions on EOB to the
               * end-of-buffer state).  Contrast this with the test
               * in input().
               */
              if (yyg->yy_c_buf_p <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars])
                { /* This was really a NUL. */
                  yy_state_type yy_next_state;

                  yyg->yy_c_buf_p = yyg->yytext_ptr + yy_amount_of_matched_text;

                  yy_current_state = yy_get_previous_state(yyscanner);

                  /* Okay, we're now positioned to make the NUL
                   * transition.  We couldn't have
                   * yy_get_previous_state() go ahead and do it
                   * for us because it doesn't know how to deal
                   * with the possibility of jamming (and we don't
                   * want to build jamming into it because then it
                   * will run more slowly).
                   */

                  yy_next_state = yy_try_NUL_trans(yy_current_state, yyscanner);

                  yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;

                  if (yy_next_state)
                    {
                      /* Consume the NUL. */
                      yy_cp = ++yyg->yy_c_buf_p;
                      yy_current_state = yy_next_state;
                      goto yy_match;
                    }

                  else
                    {
                      yy_cp = yyg->yy_c_buf_p;
                      goto yy_find_action;
                    }
                }

              else
                switch (yy_get_next_buffer(yyscanner))
                  {
                  case EOB_ACT_END_OF_FILE:
                    {
                      yyg->yy_did_buffer_switch_on_eof = 0;

                      if (yywrap(yyscanner))
                        {
                          /* Note: because we've taken care in
                           * yy_get_next_buffer() to have set up
                           * yytext, we can now set up
                           * yy_c_buf_p so that if some total
                           * hoser (like flex itself) wants to
                           * call the scanner after we return the
                           * YY_NULL, it'll still work - another
                           * YY_NULL will get returned.
                           */
                          yyg->yy_c_buf_p = yyg->yytext_ptr + YY_MORE_ADJ;

                          yy_act = YY_STATE_EOF(YY_START);
                          goto do_action;
                        }

                      else
                        {
                          if (!yyg->yy_did_buffer_switch_on_eof) YY_NEW_FILE;
                        }
                      break;
                    }

                  case EOB_ACT_CONTINUE_SCAN:
                    yyg->yy_c_buf_p = yyg->yytext_ptr + yy_amount_of_matched_text;

                    yy_current_state = yy_get_previous_state(yyscanner);

                    yy_cp = yyg->yy_c_buf_p;
                    yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
                    goto yy_match;

                  case EOB_ACT_LAST_MATCH:
                    yyg->yy_c_buf_p = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars];

                    yy_current_state = yy_get_previous_state(yyscanner);

                    yy_cp = yyg->yy_c_buf_p;
                    yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
                    goto yy_find_action;
                  }
              break;
            }

          default: YY_FATAL_ERROR("fatal flex scanner internal error--no action found");
          } /* end of action switch */
      }     /* end of scanning one token */
  }         /* end of user's declarations */
} /* end of yylex */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
static int
yy_get_next_buffer(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
  char *source = yyg->yytext_ptr;
  int number_to_move, i;
  int ret_val;

  if (yyg->yy_c_buf_p > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1])
    YY_FATAL_ERROR("fatal flex scanner internal error--end of buffer missed");

  if (YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0)
    { /* Don't try to fill the buffer, so this is an EOF. */
      if (yyg->yy_c_buf_p - yyg->yytext_ptr - YY_MORE_ADJ == 1)
        {
          /* We matched a single character, the EOB, so
           * treat this as a final EOF.
           */
          return EOB_ACT_END_OF_FILE;
        }

      else
        {
          /* We matched some text prior to the EOB, first
           * process it.
           */
          return EOB_ACT_LAST_MATCH;
        }
    }

  /* Try to read more data. */

  /* First move last chars to start of buffer. */
  number_to_move = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr - 1);

  for (i = 0; i < number_to_move; ++i) *(dest++) = *(source++);

  if (YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING)
    /* don't do the read, it's not guaranteed to return an EOF,
     * just force an EOF
     */
    YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars = 0;

  else
    {
      int num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

      while (num_to_read <= 0)
        { /* Not enough room in the buffer - grow it. */

          YY_FATAL_ERROR("input buffer overflow, can't enlarge buffer because scanner uses REJECT");
        }

      if (num_to_read > YY_READ_BUF_SIZE) num_to_read = YY_READ_BUF_SIZE;

      /* Read in more data. */
      YY_INPUT((&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]), yyg->yy_n_chars, num_to_read);

      YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
    }

  if (yyg->yy_n_chars == 0)
    {
      if (number_to_move == YY_MORE_ADJ)
        {
          ret_val = EOB_ACT_END_OF_FILE;
          yyrestart(yyin, yyscanner);
        }

      else
        {
          ret_val = EOB_ACT_LAST_MATCH;
          YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_EOF_PENDING;
        }
    }

  else
    ret_val = EOB_ACT_CONTINUE_SCAN;

  if ((yyg->yy_n_chars + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size)
    {
      /* Extend the array by 50%, plus the number we really need. */
      int new_size = yyg->yy_n_chars + number_to_move + (yyg->yy_n_chars >> 1);
      YY_CURRENT_BUFFER_LVALUE->yy_ch_buf
          = (char *) yyrealloc((void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf, (yy_size_t) new_size, yyscanner);
      if (!YY_CURRENT_BUFFER_LVALUE->yy_ch_buf) YY_FATAL_ERROR("out of dynamic memory in yy_get_next_buffer()");
      /* "- 2" to take care of EOB's */
      YY_CURRENT_BUFFER_LVALUE->yy_buf_size = (int) (new_size - 2);
    }

  yyg->yy_n_chars += number_to_move;
  YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] = YY_END_OF_BUFFER_CHAR;
  YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;

  yyg->yytext_ptr = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

  return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

static yy_state_type
yy_get_previous_state(yyscan_t yyscanner)
{
  yy_state_type yy_current_state;
  char *yy_cp;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  yy_current_state = yyg->yy_start;

  yyg->yy_state_ptr = yyg->yy_state_buf;
  *yyg->yy_state_ptr++ = yy_current_state;

  for (yy_cp = yyg->yytext_ptr + YY_MORE_ADJ; yy_cp < yyg->yy_c_buf_p; ++yy_cp)
    {
      YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 1);
      while (yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state)
        {
          yy_current_state = (int) yy_def[yy_current_state];
          if (yy_current_state >= 69) yy_c = yy_meta[yy_c];
        }
      yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
      *yyg->yy_state_ptr++ = yy_current_state;
    }

  return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
static yy_state_type
yy_try_NUL_trans(yy_state_type yy_current_state, yyscan_t yyscanner)
{
  int yy_is_jam;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner; /* This var may be unused depending upon options. */

  YY_CHAR yy_c = 1;
  while (yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state)
    {
      yy_current_state = (int) yy_def[yy_current_state];
      if (yy_current_state >= 69) yy_c = yy_meta[yy_c];
    }
  yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
  yy_is_jam = (yy_current_state == 68);
  if (!yy_is_jam) *yyg->yy_state_ptr++ = yy_current_state;

  (void) yyg;
  return yy_is_jam ? 0 : yy_current_state;
}

#ifndef YY_NO_UNPUT

#endif

#ifndef YY_NO_INPUT
#ifdef __cplusplus
static int
yyinput(yyscan_t yyscanner)
#else
static int
input(yyscan_t yyscanner)
#endif

{
  int c;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  *yyg->yy_c_buf_p = yyg->yy_hold_char;

  if (*yyg->yy_c_buf_p == YY_END_OF_BUFFER_CHAR)
    {
      /* yy_c_buf_p now points to the character we want to return.
       * If this occurs *before* the EOB characters, then it's a
       * valid NUL; if not, then we've hit the end of the buffer.
       */
      if (yyg->yy_c_buf_p < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars]) /* This was really a NUL. */
        *yyg->yy_c_buf_p = '\0';

      else
        { /* need more input */
          int offset = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr);
          ++yyg->yy_c_buf_p;

          switch (yy_get_next_buffer(yyscanner))
            {
            case EOB_ACT_LAST_MATCH:
              /* This happens because yy_g_n_b()
               * sees that we've accumulated a
               * token and flags that we need to
               * try matching the token before
               * proceeding.  But for input(),
               * there's no matching to consider.
               * So convert the EOB_ACT_LAST_MATCH
               * to EOB_ACT_END_OF_FILE.
               */

              /* Reset buffer status. */
              yyrestart(yyin, yyscanner);

              /*FALLTHROUGH*/

            case EOB_ACT_END_OF_FILE:
              {
                if (yywrap(yyscanner)) return 0;

                if (!yyg->yy_did_buffer_switch_on_eof) YY_NEW_FILE;
#ifdef __cplusplus
                return yyinput(yyscanner);
#else
                return input(yyscanner);
#endif
              }

            case EOB_ACT_CONTINUE_SCAN: yyg->yy_c_buf_p = yyg->yytext_ptr + offset; break;
            }
        }
    }

  c = *(unsigned char *) yyg->yy_c_buf_p; /* cast for 8-bit char's */
  *yyg->yy_c_buf_p = '\0';                /* preserve yytext */
  yyg->yy_hold_char = *++yyg->yy_c_buf_p;

  return c;
}
#endif /* ifndef YY_NO_INPUT */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * @param yyscanner The scanner object.
 * @note This function does not reset the start condition to @c INITIAL .
 */
void
yyrestart(FILE *input_file, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  if (!YY_CURRENT_BUFFER)
    {
      yyensure_buffer_stack(yyscanner);
      YY_CURRENT_BUFFER_LVALUE = yy_create_buffer(yyin, YY_BUF_SIZE, yyscanner);
    }

  yy_init_buffer(YY_CURRENT_BUFFER, input_file, yyscanner);
  yy_load_buffer_state(yyscanner);
}

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * @param yyscanner The scanner object.
 */
void
yy_switch_to_buffer(YY_BUFFER_STATE new_buffer, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  /* TODO. We should be able to replace this entire function body
   * with
   *		yypop_buffer_state();
   *		yypush_buffer_state(new_buffer);
   */
  yyensure_buffer_stack(yyscanner);
  if (YY_CURRENT_BUFFER == new_buffer) return;

  if (YY_CURRENT_BUFFER)
    {
      /* Flush out information for old buffer. */
      *yyg->yy_c_buf_p = yyg->yy_hold_char;
      YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
      YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
    }

  YY_CURRENT_BUFFER_LVALUE = new_buffer;
  yy_load_buffer_state(yyscanner);

  /* We don't actually know whether we did this switch during
   * EOF (yywrap()) processing, but the only time this flag
   * is looked at is after yywrap() is called, so it's safe
   * to go ahead and always set it.
   */
  yyg->yy_did_buffer_switch_on_eof = 1;
}

static void
yy_load_buffer_state(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
  yyg->yytext_ptr = yyg->yy_c_buf_p = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
  yyin = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
  yyg->yy_hold_char = *yyg->yy_c_buf_p;
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * @param yyscanner The scanner object.
 * @return the allocated buffer state.
 */
YY_BUFFER_STATE
yy_create_buffer(FILE *file, int size, yyscan_t yyscanner)
{
  YY_BUFFER_STATE b;

  b = (YY_BUFFER_STATE) yyalloc(sizeof(struct yy_buffer_state), yyscanner);
  if (!b) YY_FATAL_ERROR("out of dynamic memory in yy_create_buffer()");

  b->yy_buf_size = size;

  /* yy_ch_buf has to be 2 characters longer than the size given because
   * we need to put in 2 end-of-buffer characters.
   */
  b->yy_ch_buf = (char *) yyalloc((yy_size_t)(b->yy_buf_size + 2), yyscanner);
  if (!b->yy_ch_buf) YY_FATAL_ERROR("out of dynamic memory in yy_create_buffer()");

  b->yy_is_our_buffer = 1;

  yy_init_buffer(b, file, yyscanner);

  return b;
}

/** Destroy the buffer.
 * @param b a buffer created with yy_create_buffer()
 * @param yyscanner The scanner object.
 */
void
yy_delete_buffer(YY_BUFFER_STATE b, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  if (!b) return;

  if (b == YY_CURRENT_BUFFER) /* Not sure if we should pop here. */
    YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) 0;

  if (b->yy_is_our_buffer) yyfree((void *) b->yy_ch_buf, yyscanner);

  yyfree((void *) b, yyscanner);
}

/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a yyrestart() or at EOF.
 */
static void
yy_init_buffer(YY_BUFFER_STATE b, FILE *file, yyscan_t yyscanner)

{
  int oerrno = errno;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  yy_flush_buffer(b, yyscanner);

  b->yy_input_file = file;
  b->yy_fill_buffer = 1;

  /* If b is the current buffer, then yy_init_buffer was _probably_
   * called from yyrestart() or through yy_get_next_buffer.
   * In that case, we don't want to reset the lineno or column.
   */
  if (b != YY_CURRENT_BUFFER)
    {
      b->yy_bs_lineno = 1;
      b->yy_bs_column = 0;
    }

  b->yy_is_interactive = file ? (isatty(fileno(file)) > 0) : 0;

  errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * @param yyscanner The scanner object.
 */
void
yy_flush_buffer(YY_BUFFER_STATE b, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  if (!b) return;

  b->yy_n_chars = 0;

  /* We always need two end-of-buffer characters.  The first causes
   * a transition to the end-of-buffer state.  The second causes
   * a jam in that state.
   */
  b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
  b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

  b->yy_buf_pos = &b->yy_ch_buf[0];

  b->yy_at_bol = 1;
  b->yy_buffer_status = YY_BUFFER_NEW;

  if (b == YY_CURRENT_BUFFER) yy_load_buffer_state(yyscanner);
}

/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  @param yyscanner The scanner object.
 */
void
yypush_buffer_state(YY_BUFFER_STATE new_buffer, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  if (new_buffer == NULL) return;

  yyensure_buffer_stack(yyscanner);

  /* This block is copied from yy_switch_to_buffer. */
  if (YY_CURRENT_BUFFER)
    {
      /* Flush out information for old buffer. */
      *yyg->yy_c_buf_p = yyg->yy_hold_char;
      YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
      YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
    }

  /* Only push if top exists. Otherwise, replace top. */
  if (YY_CURRENT_BUFFER) yyg->yy_buffer_stack_top++;
  YY_CURRENT_BUFFER_LVALUE = new_buffer;

  /* copied from yy_switch_to_buffer. */
  yy_load_buffer_state(yyscanner);
  yyg->yy_did_buffer_switch_on_eof = 1;
}

/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  @param yyscanner The scanner object.
 */
void
yypop_buffer_state(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  if (!YY_CURRENT_BUFFER) return;

  yy_delete_buffer(YY_CURRENT_BUFFER, yyscanner);
  YY_CURRENT_BUFFER_LVALUE = NULL;
  if (yyg->yy_buffer_stack_top > 0) --yyg->yy_buffer_stack_top;

  if (YY_CURRENT_BUFFER)
    {
      yy_load_buffer_state(yyscanner);
      yyg->yy_did_buffer_switch_on_eof = 1;
    }
}

/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
static void
yyensure_buffer_stack(yyscan_t yyscanner)
{
  yy_size_t num_to_alloc;
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  if (!yyg->yy_buffer_stack)
    {

      /* First allocation is just for 2 elements, since we don't know if this
       * scanner will even need a stack. We use 2 instead of 1 to avoid an
       * immediate realloc on the next call.
       */
      num_to_alloc = 1; /* After all that talk, this was set to 1 anyways... */
      yyg->yy_buffer_stack = (struct yy_buffer_state **) yyalloc(num_to_alloc * sizeof(struct yy_buffer_state *), yyscanner);
      if (!yyg->yy_buffer_stack) YY_FATAL_ERROR("out of dynamic memory in yyensure_buffer_stack()");

      memset(yyg->yy_buffer_stack, 0, num_to_alloc * sizeof(struct yy_buffer_state *));

      yyg->yy_buffer_stack_max = num_to_alloc;
      yyg->yy_buffer_stack_top = 0;
      return;
    }

  if (yyg->yy_buffer_stack_top >= (yyg->yy_buffer_stack_max) - 1)
    {

      /* Increase the buffer to prepare for a possible push. */
      yy_size_t grow_size = 8 /* arbitrary grow size */;

      num_to_alloc = yyg->yy_buffer_stack_max + grow_size;
      yyg->yy_buffer_stack
          = (struct yy_buffer_state **) yyrealloc(yyg->yy_buffer_stack, num_to_alloc * sizeof(struct yy_buffer_state *), yyscanner);
      if (!yyg->yy_buffer_stack) YY_FATAL_ERROR("out of dynamic memory in yyensure_buffer_stack()");

      /* zero only the new slots.*/
      memset(yyg->yy_buffer_stack + yyg->yy_buffer_stack_max, 0, grow_size * sizeof(struct yy_buffer_state *));
      yyg->yy_buffer_stack_max = num_to_alloc;
    }
}

/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE
yy_scan_buffer(char *base, yy_size_t size, yyscan_t yyscanner)
{
  YY_BUFFER_STATE b;

  if (size < 2 || base[size - 2] != YY_END_OF_BUFFER_CHAR || base[size - 1] != YY_END_OF_BUFFER_CHAR)
    /* They forgot to leave room for the EOB's. */
    return NULL;

  b = (YY_BUFFER_STATE) yyalloc(sizeof(struct yy_buffer_state), yyscanner);
  if (!b) YY_FATAL_ERROR("out of dynamic memory in yy_scan_buffer()");

  b->yy_buf_size = (int) (size - 2); /* "- 2" to take care of EOB's */
  b->yy_buf_pos = b->yy_ch_buf = base;
  b->yy_is_our_buffer = 0;
  b->yy_input_file = NULL;
  b->yy_n_chars = b->yy_buf_size;
  b->yy_is_interactive = 0;
  b->yy_at_bol = 1;
  b->yy_fill_buffer = 0;
  b->yy_buffer_status = YY_BUFFER_NEW;

  yy_switch_to_buffer(b, yyscanner);

  return b;
}

/** Setup the input buffer state to scan a string. The next call to yylex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       yy_scan_bytes() instead.
 */
YY_BUFFER_STATE
yy_scan_string(const char *yystr, yyscan_t yyscanner) { return yy_scan_bytes(yystr, (int) strlen(yystr), yyscanner); }

/** Setup the input buffer state to scan the given bytes. The next call to yylex() will
 * scan from a @e copy of @a bytes.
 * @param yybytes the byte buffer to scan
 * @param _yybytes_len the number of bytes in the buffer pointed to by @a bytes.
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE
yy_scan_bytes(const char *yybytes, int _yybytes_len, yyscan_t yyscanner)
{
  YY_BUFFER_STATE b;
  char *buf;
  yy_size_t n;
  int i;

  /* Get memory for full buffer, including space for trailing EOB's. */
  n = (yy_size_t)(_yybytes_len + 2);
  buf = (char *) yyalloc(n, yyscanner);
  if (!buf) YY_FATAL_ERROR("out of dynamic memory in yy_scan_bytes()");

  for (i = 0; i < _yybytes_len; ++i) buf[i] = yybytes[i];

  buf[_yybytes_len] = buf[_yybytes_len + 1] = YY_END_OF_BUFFER_CHAR;

  b = yy_scan_buffer(buf, n, yyscanner);
  if (!b) YY_FATAL_ERROR("bad buffer in yy_scan_bytes()");

  /* It's okay to grow etc. this buffer, and we should throw it
   * away when we're done.
   */
  b->yy_is_our_buffer = 1;

  return b;
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

static void yynoreturn
yy_fatal_error(const char *msg, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  (void) yyg;
  fprintf(stderr, "%s\n", msg);
  exit(YY_EXIT_FAILURE);
}

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n)                                  \
  do                                               \
    {                                              \
      /* Undo effects of setting up yytext. */     \
      int yyless_macro_arg = (n);                  \
      YY_LESS_LINENO(yyless_macro_arg);            \
      yytext[yyleng] = yyg->yy_hold_char;          \
      yyg->yy_c_buf_p = yytext + yyless_macro_arg; \
      yyg->yy_hold_char = *yyg->yy_c_buf_p;        \
      *yyg->yy_c_buf_p = '\0';                     \
      yyleng = yyless_macro_arg;                   \
    }                                              \
  while (0)

/* Accessor  methods (get/set functions) to struct members. */

/** Get the user-defined data for this scanner.
 * @param yyscanner The scanner object.
 */
YY_EXTRA_TYPE
yyget_extra(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yyextra;
}

/** Get the current line number.
 * @param yyscanner The scanner object.
 */
int
yyget_lineno(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  if (!YY_CURRENT_BUFFER) return 0;

  return yylineno;
}

/** Get the current column number.
 * @param yyscanner The scanner object.
 */
int
yyget_column(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  if (!YY_CURRENT_BUFFER) return 0;

  return yycolumn;
}

/** Get the input stream.
 * @param yyscanner The scanner object.
 */
FILE *
yyget_in(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yyin;
}

/** Get the output stream.
 * @param yyscanner The scanner object.
 */
FILE *
yyget_out(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yyout;
}

/** Get the length of the current token.
 * @param yyscanner The scanner object.
 */
int
yyget_leng(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yyleng;
}

/** Get the current token.
 * @param yyscanner The scanner object.
 */

char *
yyget_text(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yytext;
}

/** Set the user-defined data. This data is never touched by the scanner.
 * @param user_defined The data to be associated with this scanner.
 * @param yyscanner The scanner object.
 */
void
yyset_extra(YY_EXTRA_TYPE user_defined, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yyextra = user_defined;
}

/** Set the current line number.
 * @param _line_number line number
 * @param yyscanner The scanner object.
 */
void
yyset_lineno(int _line_number, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  /* lineno is only valid if an input buffer exists. */
  if (!YY_CURRENT_BUFFER) YY_FATAL_ERROR("yyset_lineno called with no buffer");

  yylineno = _line_number;
}

/** Set the current column.
 * @param _column_no column number
 * @param yyscanner The scanner object.
 */
void
yyset_column(int _column_no, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  /* column is only valid if an input buffer exists. */
  if (!YY_CURRENT_BUFFER) YY_FATAL_ERROR("yyset_column called with no buffer");

  yycolumn = _column_no;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param _in_str A readable stream.
 * @param yyscanner The scanner object.
 * @see yy_switch_to_buffer
 */
void
yyset_in(FILE *_in_str, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yyin = _in_str;
}

void
yyset_out(FILE *_out_str, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yyout = _out_str;
}

int
yyget_debug(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yy_flex_debug;
}

void
yyset_debug(int _bdebug, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yy_flex_debug = _bdebug;
}

/* Accessor methods for yylval and yylloc */

YYSTYPE *
yyget_lval(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  return yylval;
}

void
yyset_lval(YYSTYPE *yylval_param, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  yylval = yylval_param;
}

/* User-visible API */

/* yylex_init is special because it creates the scanner itself, so it is
 * the ONLY reentrant function that doesn't take the scanner as the last argument.
 * That's why we explicitly handle the declaration, instead of using our macros.
 */
int
yylex_init(yyscan_t *ptr_yy_globals)
{
  if (ptr_yy_globals == NULL)
    {
      errno = EINVAL;
      return 1;
    }

  *ptr_yy_globals = (yyscan_t) yyalloc(sizeof(struct yyguts_t), NULL);

  if (*ptr_yy_globals == NULL)
    {
      errno = ENOMEM;
      return 1;
    }

  /* By setting to 0xAA, we expose bugs in yy_init_globals. Leave at 0x00 for releases. */
  memset(*ptr_yy_globals, 0x00, sizeof(struct yyguts_t));

  return yy_init_globals(*ptr_yy_globals);
}

/* yylex_init_extra has the same functionality as yylex_init, but follows the
 * convention of taking the scanner as the last argument. Note however, that
 * this is a *pointer* to a scanner, as it will be allocated by this call (and
 * is the reason, too, why this function also must handle its own declaration).
 * The user defined value in the first argument will be available to yyalloc in
 * the yyextra field.
 */
int
yylex_init_extra(YY_EXTRA_TYPE yy_user_defined, yyscan_t *ptr_yy_globals)
{
  struct yyguts_t dummy_yyguts;

  yyset_extra(yy_user_defined, &dummy_yyguts);

  if (ptr_yy_globals == NULL)
    {
      errno = EINVAL;
      return 1;
    }

  *ptr_yy_globals = (yyscan_t) yyalloc(sizeof(struct yyguts_t), &dummy_yyguts);

  if (*ptr_yy_globals == NULL)
    {
      errno = ENOMEM;
      return 1;
    }

  /* By setting to 0xAA, we expose bugs in
  yy_init_globals. Leave at 0x00 for releases. */
  memset(*ptr_yy_globals, 0x00, sizeof(struct yyguts_t));

  yyset_extra(yy_user_defined, *ptr_yy_globals);

  return yy_init_globals(*ptr_yy_globals);
}

static int
yy_init_globals(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  /* Initialization is the same as for the non-reentrant scanner.
   * This function is called from yylex_destroy(), so don't allocate here.
   */

  yyg->yy_buffer_stack = NULL;
  yyg->yy_buffer_stack_top = 0;
  yyg->yy_buffer_stack_max = 0;
  yyg->yy_c_buf_p = NULL;
  yyg->yy_init = 0;
  yyg->yy_start = 0;

  yyg->yy_start_stack_ptr = 0;
  yyg->yy_start_stack_depth = 0;
  yyg->yy_start_stack = NULL;

  yyg->yy_state_buf = 0;
  yyg->yy_state_ptr = 0;
  yyg->yy_full_match = 0;
  yyg->yy_lp = 0;

/* Defined in main.c */
#ifdef YY_STDINIT
  yyin = stdin;
  yyout = stdout;
#else
  yyin = NULL;
  yyout = NULL;
#endif

  /* For future reference: Set errno on error, since we are called by
   * yylex_init()
   */
  return 0;
}

/* yylex_destroy is for both reentrant and non-reentrant scanners. */
int
yylex_destroy(yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;

  /* Pop the buffer stack, destroying each element. */
  while (YY_CURRENT_BUFFER)
    {
      yy_delete_buffer(YY_CURRENT_BUFFER, yyscanner);
      YY_CURRENT_BUFFER_LVALUE = NULL;
      yypop_buffer_state(yyscanner);
    }

  /* Destroy the stack itself. */
  yyfree(yyg->yy_buffer_stack, yyscanner);
  yyg->yy_buffer_stack = NULL;

  /* Destroy the start condition stack. */
  yyfree(yyg->yy_start_stack, yyscanner);
  yyg->yy_start_stack = NULL;

  yyfree(yyg->yy_state_buf, yyscanner);
  yyg->yy_state_buf = NULL;

  /* Reset the globals. This is important in a non-reentrant scanner so the next time
   * yylex() is called, initialization will occur. */
  yy_init_globals(yyscanner);

  /* Destroy the main struct (reentrant only). */
  yyfree(yyscanner, yyscanner);
  yyscanner = NULL;
  return 0;
}

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void
yy_flex_strncpy(char *s1, const char *s2, int n, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  (void) yyg;

  int i;
  for (i = 0; i < n; ++i) s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int
yy_flex_strlen(const char *s, yyscan_t yyscanner)
{
  int n;
  for (n = 0; s[n]; ++n)
    ;

  return n;
}
#endif

void *
yyalloc(yy_size_t size, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  (void) yyg;
  return malloc(size);
}

void *
yyrealloc(void *ptr, yy_size_t size, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  (void) yyg;

  /* The cast to (char *) in the following accommodates both
   * implementations that use char* generic pointers, and those
   * that use void* generic pointers.  It works with the latter
   * because both ANSI C and C++ allow castless assignment from
   * any pointer type to void*, and deal with argument conversions
   * as though doing an assignment.
   */
  return realloc(ptr, size);
}

void
yyfree(void *ptr, yyscan_t yyscanner)
{
  struct yyguts_t *yyg = (struct yyguts_t *) yyscanner;
  (void) yyg;
  free((char *) ptr); /* see yyrealloc() for (char *) cast */
}

#define YYTABLES_NAME "yytables"

#line 105 "expr_lex.l"
