/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2021 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/
#ifndef UTIL_FILES_H
#define UTIL_FILES_H

#include <sys/types.h>
#include <string>
namespace FileUtils
{
bool file_exists(const std::string &filename);
bool file_exists(const char *filename);
bool user_file_overwrite(const char *filename);
off_t size(const char *filename);
void gen_suffix(char *filesuffix, size_t maxlen, int filetype, int vlistID, const char *refname);
}  // namespace FileUtils

#endif
