/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/
#ifndef GRID_PROJ_H
#define GRID_PROJ_H

#include <cstddef>

int cdo_lonlat_to_lcc(int gridID, size_t nvals, double *xvals, double *yvals);
int cdo_lcc_to_lonlat(int gridID, size_t nvals, double *xvals, double *yvals);
int cdo_stere_to_lonlat(int gridID, size_t nvals, double *xvals, double *yvals);

void cdo_sinu_to_lonlat(size_t nvals, double *xvals, double *yvals);
void cdo_laea_to_lonlat(int gridID, size_t nvals, double *xvals, double *yvals);

void cdo_proj_to_lonlat(char *proj_param, size_t nvals, double *xvals, double *yvals);

int proj_lonlat_to_lcc(struct CDI_GridProjParams gpp, size_t nvals, double *xvals, double *yvals);
int proj_lcc_to_lonlat(struct CDI_GridProjParams *gpp, double x_0, double y_0, size_t nvals, double *xvals, double *yvals);

int proj_lonlat_to_stere(struct CDI_GridProjParams gpp, size_t nvals, double *xvals, double *yvals);
int proj_stere_to_lonlat(struct CDI_GridProjParams *gpp, double x_0, double y_0, size_t nvals, double *xvals, double *yvals);

void grid_def_params_laea(int gridID, double a, double lon0, double lat0);
void grid_def_params_sinu(int gridID);

#endif /* GRID_PROJ_H */
